﻿/*
 * Author:       Michael Meijer
 * Last changed: November 18, 2012
 */

using System.Collections.Generic;
using System.Runtime.Serialization;
using Microsoft.ComplexEventProcessing;
using Microsoft.ComplexEventProcessing.Extensibility;

namespace Msdn.StreamInsight
{
    /// <summary>
    /// An user defined stream operator incrementally counting the number of 1's in a bit/binary 
    /// stream of point in time events over the last n events seen. It maintains an approximate 
    /// count with a bounded relative error upper bound.
    /// 
    /// It is a stateful operator, <see cref="IsEmpty"/>. It keeps an <see cref="ExponentialHistogram"/> 
    /// for the lifetime of the operator isntance in a query on a stream.
    /// </summary>
    [DataContract]
    public class ApproximateCountUDSO : CepPointStreamOperator<bool, long>
    {
        [DataMember]
        private ExponentialHistogram histogram;

        [DataMember]
        private long currentTimestamp;  // current (discrete) timestamp

        /// <summary>
        /// Initializes the operator with the given window size and 
        /// relative error upper bound.
        /// </summary>
        /// <param name="n">Window size</param>
        /// <param name="epsilon">Relative error upper bound</param>
        public ApproximateCountUDSO(long n, double epsilon)
        {
            // initialize a new histogram with given window size and relative error upper bound
            histogram = new ExponentialHistogram(n, epsilon);
        }

        /// <summary>
        /// Processes events as they arrive, hence incremental. Implemented 
        /// with the yield-return iterator syntax. The operator processes 
        /// only the event's payload; ignoring timestamps etc. offered by the 
        /// events 
        /// </summary>
        /// <param name="inputEvent"></param>
        /// <returns></returns>
        public override IEnumerable<long> ProcessEvent(PointEvent<bool> inputEvent)
        {
            // increment current (discrete) timestamp
            currentTimestamp++;

            // update histogram with event's payload, either 1/True or 0/False
            histogram.Update(currentTimestamp, inputEvent.Payload);

            // return approximate count over last n events
            yield return histogram.Query();
        }

        /// <summary>
        /// Indicates this operator has non-empty state, hence its state cannot be discarded.
        /// </summary>
        public override bool IsEmpty
        {
            get { return false; }
        }
    }
}
