﻿/*
 * Author:       Michael Meijer
 * Last changed: November 18, 2012
 */

using System;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.Text;
using System.Diagnostics;

namespace Msdn.StreamInsight
{
    /// <summary>
    /// An exponential histogram for approximating counts of the number of 1's over the last N 
    /// events after being presented a sequence of 0's and 1's. The approximated count is 
    /// bounded above by the relative error epsilon w.r.t. the exact count.
    /// 
    /// The histogram maintains buckets with a count and timestamp. Buckets covering only expired 
    /// events are removed. A new 1-event is put into a bucket and stored in the histogram. 
    /// Buckets are merged as much as possible to reduce memory space usage and processing time.
    /// However the oldest / last bucket covers at least one non-expired event but possible some 
    /// expired events. Hence its count is estimated. This causes the error in the estimation. 
    /// The goal of the histogram maintenance is to keep the last bucket's influence on the count 
    /// small enough.
    /// 
    /// Note 1: this implementation is not thread-safe.
    /// Note 2: the algorithm has a practical lower bound of 0.5 and 1.0 upper bound for epsilon.
    /// 
    /// Based on: "Maintaining Stream Statistics over Sliding Windows" (2002)
    ///           Datar M., Gionis, A., Indyk, P. & Motwani R.
    ///           SIAM Journal on Computing, 31(6), 1794–1813.
    ///             
    /// </summary>
    [DataContract]
    public class ExponentialHistogram
    {
        /// <summary>
        /// A bucket with a timestamp and a count. Used extensively 
        /// by the histogram <see cref="ExponentialHistogram"/>.
        /// </summary>
        [DataContract]
        private class Bucket
        {
            [DataMember]
            private long timestamp;

            [DataMember]
            private long count;

            /// <summary>
            /// The discrete timestamp of the most recent event covered by this bucket.
            /// </summary>
            public long Timestamp { get { return timestamp; } set { timestamp = value; } }

            /// <summary>
            /// The count of events covered by this bucket.
            /// </summary>
            public long Count { get { return count; } set { count = value; } }

            /// <summary>
            /// Returns string representation of the internal state of the bucket.
            /// </summary>
            /// <returns></returns>
            public override string ToString()
            {
                return string.Format("T = {0}, C = {1}", Timestamp, Count);
            }
        }

        [DataMember]
        private long n;         // window size

        [DataMember]
        private double epsilon; // relative error upper bound

        [DataMember]
        private long total;     // total count of all buckets

        // buckets in timestamp descending and count ascending order
        // the last bucket might count expired events, however it always 
        // counts for at least one non-expired event
        [DataMember]
        private LinkedList<Bucket> buckets; 

        /// <summary>
        /// Initializes the histogram for a given window size <paramref name="n"/> and 
        /// a relative error upper bound <paramref name="epsilon"/>.
        /// </summary>
        /// <param name="n">Window size</param>
        /// <param name="epsilon">Relative error upper bound</param>
        public ExponentialHistogram(long n, double epsilon)
        {
            this.n = n;
            this.epsilon = epsilon;
            //this.total = 0;
            this.buckets = new LinkedList<Bucket>();
        }

        /// <summary>
        /// Number of buckets in the histogram.
        /// </summary>
        public long BucketCount { get { return buckets.Count; } }

        /// <summary>
        /// Updates the histogram by accounting for expired events and 
        /// including the new event <paramref name="eventPayload"/>.
        /// </summary>
        /// <param name="timestamp">New event timestamp.</param>
        /// <param name="eventPayload">New event (payload) to account for.</param>
        public void Update(long timestamp, bool eventPayload)
        {
            RemoveExpiredBuckets(timestamp);

            // no new bucket required, hence done processing
            if (!eventPayload)
                return;

            PrependNewBucket(timestamp);
            
            MergeBuckets();
        }

        /// <summary>
        /// Expires buckets from <see cref="buckets"/> whose timestamp is outside the window.
        /// </summary>
        /// <param name="timestamp">New event timestamp; the current timestamp.</param>
        protected void RemoveExpiredBuckets(long timestamp)
        {
            // remove expired buckets: bucket timestamp <= current timestamp - window size
            // a bucket expires if its timestamp is not in (current timestamp - n, current timestamp]
            LinkedListNode<Bucket> node = buckets.Last;
            while (node != null && node.Value.Timestamp <= timestamp - n)
            {
                // bookkeeping: total (account for expired bucket)
                total -= node.Value.Count;

                // remove bucket
                buckets.RemoveLast();
                node = buckets.Last;
            }
        }

        /// <summary>
        /// Prepends a new bucket in <see cref="buckets"/> to account for a 1-event.
        /// </summary>
        /// <param name="timestamp">New bucket timestamp, which equals the new event timestamp it covers.</param>
        protected void PrependNewBucket(long timestamp)
        {
            // a new bucket with current timestamp and count 1
            Bucket newBucket = new Bucket()
            {
                Timestamp = timestamp,
                Count = 1
            };

            // prepend to maintain timestamp descending and count ascending order
            buckets.AddFirst(newBucket);

            // bookkeeping: total (account for new bucket)
            total++;
        }

        /// <summary>
        /// Merges buckets in <see cref="buckets"/>. Whenever k/2+2 buckets with the same count are found, the 
        /// oldest two of those buckets are merged into a single bucket. The merging can cascade from thereon.
        /// </summary>
        protected void MergeBuckets()
        {
            LinkedListNode<Bucket> current = buckets.First;
            LinkedListNode<Bucket> previous = null;

            int k = (int)Math.Ceiling(1 / epsilon);             // k=1/eps as integer
            int kDiv2Add2 = (int)(Math.Ceiling(0.5 * k) + 2);   // k/2 as integer
            // at this point 1/k <= eps, k >= 2, hence requires eps >= 0.5

            // number of concecutive buckets with same count causing a 
            // merge of the oldest two of those buckets
            int numberOfSameCount = 0;

            // traverse buckets from first to last, hence in order of 
            // descending timestamp and ascending count
            while (current != null)
            {
                // previous and current bucket have same count, increment counter
                if (previous != null && previous.Value.Count == current.Value.Count)
                    numberOfSameCount++;
                // current is first with that count, reset counter to 1
                else
                    numberOfSameCount = 1;

                // detect need for a merge
                if (numberOfSameCount == kDiv2Add2)
                {
                    // merge buckets into current and remove previous
                    current.Value.Timestamp = previous.Value.Timestamp;                 // take most recent timestamp
                    current.Value.Count = previous.Value.Count + current.Value.Count;   // sum the counts of the buckets, 
                                                                                        // i.e. next power of two

                    buckets.Remove(previous);

                    // note that a merged bucket might cause a cascade of merges due to its new count,
                    // hence the new current node should point to the merged bucket otherwise the 
                    // cascade might go unnoticed, temporarily violating the invariant!

                    previous = current.Previous;    // merged bucket's previous, since old previous is removed
                    //current = current;            // trivial, merged bucket is new current

                    // at this iteration, the traversal stays in place
                }
                // no merge required, continue normally
                else
                {
                    previous = current;         // old current bucket or merged bucket
                    current = current.Next;     // current's or merged's next

                    // at this iteration, the traversal moves to the next (older) bucket
                }
            }
        }

        /// <summary>
        /// Queries the histogram for the approximate count with a relative error of 
        /// at most epsilon.
        /// </summary>
        /// <returns></returns>
        public long Query()
        {
            /*
             * The sum of all bucket counts up to, exclusive, the last bucket are 
             * exact counts. The last bucket count is estimated by half its count.
             * This ensures an absolute error no larger than half that bucket's 
             * count.
             * Assuming proper maintenance by Update(bool) the resulting estimate 
             * should be within the relative error upper bound epsilon.
             */
            long last = buckets.Last != null ? buckets.Last.Value.Count : 0;

            return (long)Math.Ceiling(total - last / 2.0);
        }

        /// <summary>
        /// Returns string representation of the internal state of the histogram.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder builder = new StringBuilder();

            int k = (int)(Math.Ceiling(0.5 * Math.Ceiling(1 / epsilon)) * 2);
            builder.Append(string.Format("EH: N = {0}, epsilon = {1}, k = {2}, #B = {3}, C = {4}\n\n", n, epsilon, k, buckets.Count, Query()));

            foreach (Bucket bucket in buckets)
                builder.Append(string.Format("\tBucket: {0}\n", bucket));

            return builder.ToString();
        }
    }

    /// <summary>
    /// Represents a window of the last N events. Based on a circular buffer implementation. 
    /// Used to compare against exponential histogram approximated counts.
    /// </summary>
    public class Window
    {
        private long n;
        private long ct;
        private long count;
        private bool[] events;
        private long i;

        public Window(long n)
        {
            this.n = n;
            //this.ct = 0;
            //this.count = 0;
            this.events = new bool[n];
            //this.i = 0;
        }

        public void Update(bool e)
        {
            ct++;

            if (ct > n && events[i])
                count--;

            events[i] = e;

            i = (i + 1) % n;

            if (e)
                count++;
        }

        public long Query()
        {
            return count;
        }
    }

    /// <summary>
    /// Demonstration of the exponential histogram.
    /// </summary>
    public static class Demonstration
    {
        /// <summary>
        /// Demonstrates the exponential histogram on a random sequence of 0's and 1's.
        /// Compares the results with an exact count.
        /// </summary>
        public static void DemoMain()
        {
            int seed = (int)DateTime.Now.Ticks;
            Random random = new Random(seed);   // event generator
            int max = 100000000;                // stream length
            Stopwatch stopwatch = new Stopwatch();
            double sumRelativeError = 0.0;      // summed relative error
            double maxRelativeError = 0.0;      // maximum relative error
            //long underestimates = 0;            // number of underestimates
            //long overestimates = 0;             // number of overestimates
            //long exacts = 0;                    // number of exacts
            long sumBucketCount = 0;            // summed count of buckets
            long maxBucketCount = 0;            // maximum count of buckets

            int n = 1000000;    // window size
            double eps = 0.01;  // 1% relative error upper bound -> 99% accuracy
            ExponentialHistogram histogram = new ExponentialHistogram(n, eps);
            Window window = new Window(n);

            for (int i = 0; i < max; i++)
            {
                // randomly generate element
                bool element = random.NextDouble() >= 0.5;

                // update and query histogram for approximate count
                stopwatch.Start();

                histogram.Update(i, element);
                long approxCount = histogram.Query();

                stopwatch.Stop();

                // update window ringbuffer
                window.Update(element);
                long exactCount = window.Query();

                // validate result: relative error of apprixmation w.r.t. exact bounded by epsilon
                // wait until first window of n events passed, then maintain stats
                if (i >= n)
                {
                    // classify current approximate count as underestimates, overestimates or exact
                    //underestimates += exactCount > approxCount ? 1 : 0;
                    //overestimates += exactCount < approxCount ? 1 : 0;
                    //exacts += exactCount == approxCount ? 1 : 0;

                    // determine relative error for currrent approximate count w.r.t. current exact count
                    if (exactCount > 0) // prevent division by zero
                    {
                        double relativeError = Math.Abs(exactCount - approxCount) / (double)exactCount;

                        // sum and max relative error observed
                        sumRelativeError += relativeError;
                        maxRelativeError = Math.Max(maxRelativeError, relativeError);
                    }

                    // sum and max bucket counts in histogram
                    sumBucketCount += histogram.BucketCount;
                    maxBucketCount = Math.Max(maxBucketCount, histogram.BucketCount);
                }
            }

            // print histogram state
            //Console.WriteLine(histogram);

            // print stats
            Console.WriteLine("Elements passed:          {0}", max);
            Console.WriteLine("Window size:              {0}", n);
            Console.WriteLine("Relative error bound:     {0}", eps);
            //Console.WriteLine("Underestimates of count:  {0}", underestimates);
            //Console.WriteLine("Overestimates of count:   {0}", overestimates);
            //Console.WriteLine("Exacts of count:          {0}", exacts);
            Console.WriteLine("Average relative error:   {0:0.#####}", sumRelativeError / max);
            Console.WriteLine("Maximum relative error:   {0:0.#####}", maxRelativeError);
            Console.WriteLine("Relative error violation: {0}", maxRelativeError > eps);
            Console.WriteLine("Average bucket count:     {0}, size relative to window {1:P}", sumBucketCount / max, (double)(sumBucketCount / max) / n);
            Console.WriteLine("Maximum bucket count:     {0}, size relative to window {1:P}", maxBucketCount, (double)maxBucketCount / n);
            Console.WriteLine("Avg time per element:     {0:0.000} microsecs / element", ((double)stopwatch.ElapsedTicks / Stopwatch.Frequency) / max * 1000000);
            Console.WriteLine("Elements per time:        {0:0} elements / second", max / ((double)stopwatch.ElapsedTicks / Stopwatch.Frequency));

            Console.WriteLine("\n<Press ENTER to quit>");
            Console.ReadLine();
        }
    }
}
