﻿/*
 * Author:       Michael Meijer
 * Last changed: November 18, 2012
 */

using System;
using Microsoft.ComplexEventProcessing;
using Microsoft.ComplexEventProcessing.Linq; // has useful extension methods

namespace Msdn.StreamInsight
{
    /// <summary>
    /// Demonstrates the use of the <see cref="ApproximateCountUDSO"/> in a StreamInsight query.
    /// </summary>
    class Program
    {
        /// <summary>
        /// A predefined window size.
        /// </summary>
        public const long N = 10000;

        /// <summary>
        /// A predefined relative error upper bound.
        /// </summary>
        public const double Epsilon = 0.05;

        static void Main(string[] args)
        {
            // use in-proc embedded StreamInsight server instance
            using (Server server = Server.Create("StreamInsight21"))
            {
                // some (metadata-)container application
                var app = server.CreateApplication("ApproximateCountDemo");

                // define an enumerable source
                var source = app.DefineEnumerable(() => Utility.EnumeratePayloads());

                // wrap the source in a (temporal) point in time event stream with increasing start times
                // the advance time settings determine when CTI events are generated by StreamInsight
                var sourceStream = source.ToPointStreamable(e => PointEvent.CreateInsert(DateTime.Now, e), AdvanceTimeSettings.IncreasingStartTime);

                // define the query: compute and select approximate counts over the source stream
                // note: the query produces a stream of approximate counts
                var query = from e in sourceStream.ApproximateCount(N, Epsilon) select e;

                // unwrap the query (temporal) point in time stream to an enumerable sink
                var sink = query.ToEnumerable<long>();

                // start pulling/enumerating results from the sink 
                // all the way through the processing pipeline up to the source
                foreach (long estimatedCount in sink)
                    Console.WriteLine(string.Format("Enumerated Approximate count: {0}", estimatedCount));
            }
        }
    }
}
